from .core import _Value


class Float(_Value):
    __slots__ = ()

    def incr(self, value=1):
        self._memoryview[0] = value

    __iadd__ = incr

    def _get_size(self):
        return 8

    def _get_type(self):
        return ('d', 'level 8 float')

    def set(self, value):
        self._memoryview[0] = value

    def __setitem__(self, key, value):
        assert key == 0, "Only single value is expected"
        self._memoryview[key] = value


class Integer(_Value):
    __slots__ = ()

    def _get_size(self):
        return 8

    def _get_type(self):
        return ('q', 'level 8 signed')

    def set(self, value):
        self._memoryview[0] = value

    def get(self):
        return self._memoryview[0]

    def incr(self, value=1):
        self._memoryview[0] += value

    def decr(self, value=1):
        self._memoryview[0] -= value

    def __setitem__(self, key, value):
        assert key == 0, "Only single value is expected"
        self._memoryview[key] = value



from .collection import Collection, DuplicateValueException
from .collection import start
from .counters import Counter
from .levels import Float, Integer
from .simple_state import State
from .fork import Fork
from .req import RequestTracker


__all__ = [
    "Collection", "DuplicateValueException",
    "Counter",
    "Fork",
    "RequestTracker",
    "Float", "Integer",  # may be better names?
    "State",
    "start",
    ]



import time
import logging
import traceback
from contextlib import contextmanager

from .simple_state import State
from .counters import Counter


log = logging.getLogger(__name__)


class Branch(object):
    __slots__ = ('name', '_parent', '_counter', '_duration')

    def __init__(self, suffix, state, parent, **kwargs):
        self.name = suffix
        self._parent = parent
        self._counter = Counter(state=state + '.' + suffix,
                                metric='count', **kwargs)
        self._duration = Counter(state=state + '.' + suffix,
                                 metric='duration', **kwargs)

    def enter(self):
        self._parent.enter_branch(self)

    def _commit(self, start, fin):
        self._counter.incr(1)
        self._duration.incr(fin - start)


class Fork(object):

    def __init__(self, branches, state, **kwargs):
        state_obj = State(state=state, **kwargs)
        for name in branches:
            setattr(self, name, Branch(name,
                                       parent=self, state=state, **kwargs))
        self._state = state_obj
        self._branch = None

    @contextmanager
    def context(self):
        if self._branch is not None:
            tb = traceback.format_stack()
            log.error("Nested Fork(%x).context() is not supported at:\n%s",
                      id(self._state), ''.join(tb[:-2]).rstrip())
        self._branch = None
        self._state.enter('_')
        try:
            yield
        finally:
            ts = int(time.time()*1000)
            if self._branch is not None:
                self._branch._commit(self._timestamp, ts)
            self._state.exit()
            self._branch = None

    def enter_branch(self, branch):
        ts = int(time.time()*1000)
        if self._branch is not None:
            self._branch._commit(self._timestamp, ts)
        self._state.enter(branch.name, _timestamp=ts)
        self._timestamp = ts
        self._branch = branch



import time
from contextlib import contextmanager

from .counters import Counter
from .levels import Integer


class RequestTracker(object):
    def __init__(self, group_name, **kwargs):
        self.requests = Counter(group=group_name,
                                metric="requests", **kwargs)
        self.duration = Counter(group=group_name,
                                metric="total_duration", **kwargs)
        self.errors = Counter(group=group_name,
                              metric="errors", **kwargs)
        self.in_progress = Integer(group=group_name,
                                   metric="in_progress", **kwargs)

    @contextmanager
    def request(self):
        start = time.time()
        self.in_progress.incr()
        try:
            yield
        except Exception:
            self.errors.incr()
            raise
        finally:
            dur = time.time() - start
            self.in_progress.decr()
            self.requests.incr()
            self.duration.incr(int(dur * 1000))



from contextlib import contextmanager
from typing import Dict, Tuple, Iterator


class RequestTracker(object):
    def __init__(self, group_name: str, **kwargs: Dict[str, str]) -> None:
        pass

    @contextmanager
    def request(self) -> Iterator[None]:
        pass



import os
import mmap
import json
import atexit
import warnings
from itertools import groupby

from .py2_compat import MemoryView


class DuplicateValueException(RuntimeError):
    """Raised when all parameters to value are duplicated"""


def _get_size(pair):
    return pair[1]._get_size()


def _counter_order(pair):
    return (pair[1]._get_size(), pair[0])


class Collection(object):
    """A collection of statistics parameters

    It's just a singleton, which is hold in this module, but we use it
    as a regular class for unittests
    """

    def __init__(self):
        self._all_values = {}

    def add(self, dimensions, value):
        name = json.dumps(dimensions, sort_keys=True)
        if name in self._all_values:
            raise DuplicateValueException(
                "Counter {} is already defined".format(name))
        self._all_values[name] = value

    def start(self, basepath):
        values = list(self._all_values.items())
        del self._all_values

        values.sort(key=_counter_order)

        offset = 0
        scheme = []
        offsets = {}
        for size, pairs in groupby(values, key=_get_size):
            if size & (size-1) == 0:  # power of two; let's optimize
                if offset % size:
                    pad = size - offset % size
                    offset += pad
                    scheme.append('pad {}'.format(pad))
            elif size % 8 == 0:
                # unless value is small or
                # it's size is crappy we must align to 8
                if offset % 8:
                    pad = size - offset % 8
                    offset += pad
                    scheme.append('pad {}'.format(pad))
            for name, value in pairs:
                offsets[value] = offset
                offset += size
                _, typ = value._get_type()
                scheme.append(typ + ': ' + name)

        size = offset

        path = basepath + '.values'
        tmppath = basepath + '.tmp'
        metapath = basepath + '.meta'

        if os.path.exists(metapath):
            os.unlink(metapath)
        if os.path.exists(path):
            os.unlink(path)
        if os.path.exists(tmppath):
            os.unlink(tmppath)

        with open(tmppath, 'w+b') as f:
            # We could use truncate, but it doesn't enlarge file in
            # cross-platform fasion. Fortunately our data is small and usually
            # in RAM anyway
            f.write(b'\x00' * offset)
            f.flush()
            mem = MemoryView(mmap.mmap(f.fileno(), offset))

        os.rename(tmppath, path)

        with open(tmppath, 'wt') as f:
            f.write('\n'.join(scheme))

        os.rename(tmppath, metapath)

        for value, offset in offsets.items():
            vtype, _ = value._get_type()
            size = value._get_size()
            value._memoryview = mem[offset:offset+size].cast(vtype)

        return ActiveCollection(basepath)


class ActiveCollection(object):

    def __init__(self, path):
        self._path = path

    def add(self, dimensions, value):
        raise RuntimeError(
            "Counters can't be added after collection.start()")

    def start(self, path):
        raise RuntimeError("The start() method already called")

    def close(self):
        os.unlink(self._path + '.meta')
        os.unlink(self._path + '.values')


global_collection = Collection()


def start():
    global global_collection

    if len(global_collection._all_values) == 0:
        warnings.warn(
            "Cantal library is activated but no metrics found. "
            "Note: you need to import your modules and initialize all metrics "
            "before calling `cantal.start()`",
            stacklevel=2)
        return

    path = os.environ.pop("CANTAL_PATH", None)
    if path is None:
        if 'XDG_RUNTIME_DIR' in os.environ:
            path = '{}/cantal.{}'.format(
                os.path.join(os.environ['XDG_RUNTIME_DIR']),
                os.getpid())
        else:
            path = '/tmp/cantal.{}.{}'.format(
                os.getuid(),
                os.getpid())
        warnings.warn(
            "No CANTAL_PATH is set in the environment, using {!r}. "
            "The cantal-agent will be unable to discover it.".format(path),
            stacklevel=2)

    global_collection = global_collection.start(path)
    atexit.register(global_collection.close)



from typing import Dict, Tuple, Iterator
from contextlib import contextmanager

from .core import _Value


class State(_Value):
    def __init__(self, size: int, **kwargs: Dict[str, str]) -> None:
        pass

    def _get_size(self) -> int: pass
    def _get_type(self) -> Tuple[str, str]: pass

    def enter(self, value:str) -> None: pass

    @contextmanager
    def context(self, value:str) -> Iterator[State]:
        pass

    def exit(self) -> None: pass



from typing import Tuple

from .core import _Value


class Float(_Value):
    __slots__ = ()

    def incr(self, value: float = 1) -> None: pass
    def _get_size(self) -> int: pass
    def _get_type(self) -> Tuple[str, str]: pass
    def set(self, value: float) -> None: pass
    def __setitem__(self, key: int, value: float) -> None: pass


class Integer(_Value):
    def _get_size(self) -> int: pass
    def _get_type(self) -> Tuple[str, str]: pass
    def set(self, value: int) -> None: pass
    def get(self) -> int: pass
    def incr(self, value:int = 1) -> None: pass
    def decr(self, value:int = 1) -> None: pass
    def __setitem__(self, key: int, value: int) -> None: pass



from typing import Dict, Union
from .core import _Value


global_collection: Union[Collection, ActiveCollection]


class DuplicateValueException(RuntimeError):
    pass


class Collection(object):
    def __init__(self) -> None: pass
    def add(self, dimensions: Dict[str, str], value: _Value) -> None: pass
    def start(self, basepath: str) -> ActiveCollection: pass


class ActiveCollection(object):

    def __init__(self, path: str) -> None: pass

    def add(self, dimensions: Dict[str, str], value: _Value) -> None:
        pass

    def start(self, path: str) -> None: pass

    def close(self) -> None: pass


def start() -> None:
    pass



from typing import Union
from ctypes import c_char, c_long, c_ulonglong, c_double
import mmap


class MemoryView:
    def __init__(self, _mmap: mmap.mmap, _slice: slice) -> None: pass
    def __getitem__(self, item: slice) -> MemoryView: pass
    def cast(self, typ: str) -> Union[c_char, c_long, c_ulonglong, c_double]:
        pass



import sys
import struct

if sys.version_info >= (3, 4, 0):
    MemoryView = memoryview
else:

    import ctypes

    TYPEMAP = {
        'B': ctypes.c_char,
        'q': ctypes.c_long,
        'Q': ctypes.c_ulonglong,
        'd': ctypes.c_double,
    }

    class _MemoryView(object):
        __slots__ = ('_mmap', '_slice')

        def __init__(self, _mmap, _slice=None):
            self._mmap = _mmap
            self._slice = _slice

        def __getitem__(self, item):
            assert isinstance(item, slice), "Only slicing is supported"
            assert self._slice is None, "Only single level of slicing required"
            return MemoryView(self._mmap, item)

        def cast(self, typ):
            slc = self._slice
            if slc is None:
                slc = slice(None)  # full slice
            start, stop, step = slc.indices(len(self._mmap))
            assert step == 1, 'Only step 1 supported'
            bytesize = struct.calcsize(typ)
            _typ = TYPEMAP[typ] * ((stop - start) // bytesize)
            return _typ.from_buffer(self._mmap, start)

    MemoryView = _MemoryView



import time
import struct
import warnings
from contextlib import contextmanager

from .core import _Value
from .const import CACHE_LINE_SIZE


_timestr = struct.Struct('Q')


class State(_Value):
    __slots__ = ('size',)
    HEADER_SIZE = _timestr.size
    assert HEADER_SIZE == 8, 'We use constants in enter/exit/context'

    def __init__(self, size=CACHE_LINE_SIZE-HEADER_SIZE, **kwargs):
        sz = size + self.HEADER_SIZE
        if sz & (sz - 1) or sz % CACHE_LINE_SIZE:
            warnings.warn(
                "Size of state counter should be multiple of {} or smaller"
                "power of two sans header size ({}), perfect size is {}"
                .format(CACHE_LINE_SIZE, self.HEADER_SIZE,
                        CACHE_LINE_SIZE - self.HEADER_SIZE))
        self.size = size
        super(State, self).__init__(**kwargs)

    def _get_size(self):
        return self.HEADER_SIZE + self.size

    def _get_type(self):
        return ('B', 'state {}'.format(self.size + self.HEADER_SIZE))

    def enter(self, value, _timestamp=None):
        """Enter a state with a value

        The _timestamp is a merely optimization, don't pass it in any user
        code
        """
        encoded = value.encode('utf-8')
        le = len(encoded)
        tail = le - self.size
        if tail < 0:
            encoded += b'\x00'
            le += 1
        elif tail > 0:
            encoded = encoded[:self.size]
        le += 8
        if _timestamp is None:
            _timestamp = int(time.time()*1000)
        chunk = _timestr.pack(_timestamp) + encoded
        self._memoryview[0:le] = chunk

    @contextmanager
    def context(self, value):
        self.enter(value)
        try:
            yield self
        finally:
            self._memoryview[0:8] = b'\x00\x00\x00\x00\x00\x00\x00\x00'

    def exit(self):
        self._memoryview[0:8] = b'\x00\x00\x00\x00\x00\x00\x00\x00'



from contextlib import contextmanager
from typing import Dict, Iterator


class Branch(object):

    def __init__(self,
        suffix: str, state:str, parent: Fork,
        **kwargs: Dict[str, str],
    ) -> None:
        pass

    def enter(self) -> None: pass


class Fork(object):

    def __init__(self,
        branches: Iterator[str], state: str,
        **kwargs: Dict[str, str],
    ) -> None:
        pass

    @contextmanager
    def context(self) -> Iterator[None]: pass

    def enter_branch(self, branch: str) -> None: pass



import abc
import logging

from . import collection as _collection
from .py2_compat import MemoryView


log = logging.getLogger(__name__)


class _Value(object):
    __slots__ = ('_memoryview',)

    def __init__(self, *args, **kwargs):
        if args:
            raise TypeError("No positional arguments expected")
        collection = kwargs.pop('collection', None)
        if collection is None:
            collection = _collection.global_collection
        collection.add(kwargs, self)

        log.debug("New %s:%x with params %r",
                  self.__class__.__name__, id(self), kwargs)

        # Create temporary view, this is useful if you enter some branch of
        # code in unusual code path such as ipython shell or management
        # commands. Counters are not useful in this case but should not crash,
        # so you don't need a lot of `if` statements
        vtype, _ = self._get_type()
        size = self._get_size()
        self._memoryview = MemoryView(bytearray(size)).cast(vtype)

    @abc.abstractmethod
    def _get_size(self):
        pass



#!/usr/bin/env python

from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), 'rt') as f:
    long_description = f.read()


setup(name='cantal',
      version='0.2.3',
      description='The library that submits metrics to cantal',
      author='Paul Colomiets',
      author_email='paul@colomiets.name',
      url='http://github.com/tailhook/cantal-py',
      packages=['cantal'],
      package_data={"cantal": ["py.typed", "*.pyi"]},
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
      ],
      long_description=long_description,
      )



from .core import _Value


class Counter(_Value):
    __slots__ = ()

    def __iadd__(self, value):
        self._memoryview[0] += value
        return self

    def incr(self, value=1):
        self._memoryview[0] += value

    def _get_size(self):
        return 8

    def _get_type(self):
        return ('Q', 'counter 8')



from .core import _Value
from typing import Tuple


class Counter(_Value):
    def __iadd__(self, value: int) -> int: pass
    def incr(self, value:int=1) -> None: pass
    def _get_size(self) -> int: pass
    def _get_type(self) -> Tuple[str, str]: pass
